from typing import *
from .abstract_prompter import IPrompter, T, Generic
from .address import Address
from .address_builder import AddressBuilderGC
import re
from uuid import uuid4


def _parse(template: str):
    parts = re.split('(<<[^>]+>>)', template)
    addresses = {}
    operators = {}
    template_parts = []
    for part in parts:
        if part.startswith('<<') and part.endswith('>>'):
            uid = part[2:-2]
            if len(uid) == 0:
                raise ValueError("Expected address or pre-generated address between << and >> ")
            if uid[0] == '#':
                addresses[uid] = AddressBuilderGC.cache[uid[1:]]
                if uid in AddressBuilderGC.operators:
                    operators[uid] = AddressBuilderGC.operators[uid]
                template_parts.append('{' + uid + '}')
            else:
                address = Address.parse(uid)
                uid = str(uuid4())
                addresses[uid] = address
                template_parts.append('{' + uid + '}')
        else:
            template_parts.append(part)

    return ''.join(template_parts), addresses, operators


class Prompter(IPrompter[T], Generic[T]):
    def __init__(self, template: str):
        self.template, self.field_to_address, self.field_to_formatter = _parse(template)

    def __call__(self, obj: T) -> str:
        values = {}
        for key, address in self.field_to_address.items():
            o = address.get(obj)
            if key in self.field_to_formatter:
                o = self.field_to_formatter[key](o)
            values[key] = o
        return self.template.format(**values)





from .abstract_prompter import IPrompter
from .address import Address, IAddressElement, DefaultElement
from .address_builder import AddressBuilder, AddressBuilderGC
from .prompter import Prompter
from .referrer import Referrer



from typing import *
from .address import DefaultElement, Address
from uuid import uuid4


class AddressBuilderGC:
    cache: dict[str, Address] = {}
    operators: dict[str, Any] = {}


RESERVED_FIELDS = {
    '_address_builder_stored_address',
    '_address_builder_uuid',
    '__str__',
    '__truediv__',
}

class AddressBuilder:
    def __init__(self, address: Address|None = None):
        self._address_builder_uuid = 'id'+str(uuid4()).replace('-','')
        self._address_builder_stored_address = address if address is not None else Address()
        AddressBuilderGC.cache[self._address_builder_uuid] = self._address_builder_stored_address

    def __str__(self):
        return f'<<#{self._address_builder_uuid}>>'

    def __getattribute__(self, item):
        if item in RESERVED_FIELDS:
            return super().__getattribute__(item)
        return AddressBuilder(self._address_builder_stored_address.append(DefaultElement(item)))

    def __truediv__(self, other):
        AddressBuilderGC.operators[self._address_builder_uuid] = other
        return self.__str__()
